/*
 * Macro / inline / normal function to add/mark "timestamps" to program
 * in a way that can be found from LTT traces.  This helps in pinpointing
 * when something interesting starts or ends in the trace. Strace can
 * be used to get timings for these measurement points.
 * 
 * Copyright (C) 2008-2009 by Nokia Corporation
 *
 * Contact: Eero Tamminen <eero.tamminen@nokia.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 * USAGE:
 *	#include "sp_timestamp.h"
 * ...
 *	sp_timestamp("something-interesting-started");
 * 
 * To enable timestamping when building the code in Scratchbox,
 * do this before build:
 *	export SBOX_EXTRA_COMPILER_ARGS="-DSP_TIMESTAMP_CREATE=1"
 * 
 * You can see the timestamps when running the software with:
 *	strace -f -ttt -e trace=clone,fork,open,execve ./binary 2>&1 | grep /tmp/stamps
 * 
 * If the programs are already running, give the program IDs which
 * strace should attach to with the "-p PID" arguments instead of
 * a program name.
 * 
 * VERSIONS:
 * - Above example uses the inline version of the timestamp code which
 *   is safe and easiest to use (doesn't require linking anything), but
 *   it has most overhead as it needs to check process name on each
 *   call (i.e. you get extra open() calls in your strace log)
 * - To use a macro version, define also SP_TIMESTAMP_MACRO.
 *   Macro is fastest, but shows source file name instead of
 *   the process name and if code where it's used, uses names
 *   of standard C function names like "close" or "open" as
 *   variable names, compiler will complain (or do funny things).
 * - If you want to use a library version, you don't need to define
 *   anything, just use the sp-timestamp pkg-config file:
 *      gcc $(pkg-config --cflags --libs sp_timestamp) mycode.c
 *
 * Implementation differs from Federico's Gnome app stracing tool:
 *     http://www.gnome.org/~federico/news-2006-03.html#timeline-tools
 * - Lower level C-library functions are used instead of Glib
 * - Gets process name from /proc/PID/cmdline instead of Gtk
 *   specific functions (macro version uses just __FILE__)
 * - Uses open() syscall instead of access() because LTT doesn't
 *   record filename used by access() and because programs do
 *   use open() less (ptrace has its own overhead)
 * - Files are in /tmp which on Maemo is tmpfs (in RAM)
 */
#ifndef SP_TIMESTAMP_H
#define SP_TIMESTAMP_H

#ifdef __cplusplus
extern "C" {
#endif

#define SP_TIMESTAMP_PATH "/tmp/stamps/"

#ifndef SP_TIMESTAMP_CREATE
# define sp_timestamp(step)
#else

# if SP_TIMESTAMP_EXTERN
extern void sp_timestamp(const char *step);

# elif SP_TIMESTAMP_MACRO
#  include <sys/stat.h>
#  include <unistd.h>
#  include <fcntl.h>

/* low overhead macro adding code filename to measurement filename */
#  define sp_timestamp(step) { int _tmp_fd; \
	_tmp_fd = open(SP_TIMESTAMP_PATH __FILE__ ":" step, O_CREAT|O_WRONLY, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH); \
	if (_tmp_fd >=0) close(_tmp_fd); }

# else /* not _EXTERN nor _MACRO -> inline (default) */

#  include <sys/stat.h>
#  include <unistd.h>
#  include <fcntl.h>
#  include <sys/types.h>
#  include <string.h>
#  include <assert.h>

/* higher overhead Linux specific function that adds program name
 */
static inline void sp_timestamp(const char *step)
{
	char filename[256] = SP_TIMESTAMP_PATH;
	int offset = sizeof(SP_TIMESTAMP_PATH)-1;
	int maxlen = sizeof(filename)-1;
	int fd, count;
	char *cmdname;

	/* get process command name */
	fd = open("/proc/self/cmdline", O_RDONLY);
	assert(fd >= 0);

	count = read(fd, filename+offset, maxlen-offset);
	filename[maxlen] = '\0';
	assert(count > 0);
	close(fd);

	/* use basename of the command name */
	cmdname = strrchr(filename+offset, '/');
	memmove(filename+offset, cmdname+1, strlen(cmdname));
	offset += strlen(filename+offset);
	
	/* add step name */
	filename[offset++] = ':';
	strncpy(filename+offset, step, maxlen-offset);
	filename[maxlen] = '\0';
	
	fd = open(filename, O_CREAT|O_WRONLY, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
	if (fd >= 0) {
		close(fd);
	}
}
# endif
#endif

#ifdef __cplusplus
}
#endif

#endif /* SP_TIMESTAMP_H */
